import { deepCloneJSON } from '@codeperate/utils';
import { computed, onMounted, Ref, ref } from 'vue';
import { FormBuilderBaseContext } from '.';
import { EmitType } from '../type/emit-type';
import { debounce, LocalStorage } from '../utils';

export interface UseSaveLoadOptions {
    saveLocationPrefix: string;
    saveLocation?: string;
    saveExpiry: number;
    saveDelay: number;
    autoSave?: boolean;
    autoLoad?: boolean;
    noExpiry?: boolean;
}

export interface UseSaveLoad {
    save: () => Promise<void>;
    load: () => Promise<void>;
    cleanHistory: () => void;
    history: Ref<FormHistory | undefined>;
}

export interface FormHistory {
    value: any;
    expiryAt: number;
    noExpiry?: boolean;
    lastEdit: number;
}

export type UseSaveLoadEmits = {
    save: [];
    load: [];
};

export type WithUseSaveLoad<T> = {
    emit: EmitType<UseSaveLoadEmits>;
} & T;

export const useSaveLoad = (
    { getValue, getWidgets, setValue: formSetValue, emit }: WithUseSaveLoad<FormBuilderBaseContext>,
    props: Partial<UseSaveLoadOptions>,
) => {
    const defaultOpts: UseSaveLoadOptions = {
        saveLocationPrefix: 'form:',
        saveExpiry: 1000 * 60 * 60 * 24,
        saveDelay: 2000,
    };
    const saveExpiry = computed(() => props.saveExpiry ?? defaultOpts.saveExpiry);
    const saveLocationPrefix = computed(() => props.saveLocationPrefix ?? defaultOpts.saveLocationPrefix);
    const saveLocation = computed(() => props.saveLocation ?? defaultOpts.saveLocation);
    const saveLocationFull = computed(() => saveLocationPrefix.value + saveLocation.value);
    const saveDelay = computed(() => props.saveDelay ?? defaultOpts.saveDelay);
    const formHistory = ref<FormHistory>();
    let skipSave = false;

    const setValue: FormBuilderBaseContext['setValue'] = (...args) => {
        formSetValue(...args);
        const [, , options] = args;
        if (props.autoSave && !options?.disableSideEffect) save();
    };

    const _save = async () => {
        if (!props.saveLocation) throw new Error('You must define save location first.');
        if (skipSave) {
            skipSave = false;
            return;
        }
        try {
            const value = deepCloneJSON(getValue());
            const widgets = getWidgets();
            const promises: Promise<any>[] = [];
            for (const widget of widgets) promises.push(widget.onSaveHistory(value));
            await Promise.all(promises);
            LocalStorage.set<FormHistory>(saveLocationFull.value, {
                value,
                expiryAt: new Date().getTime() + saveExpiry.value,
                lastEdit: new Date().getTime(),
                noExpiry: props.noExpiry,
            });
            emit('save');
        } catch (err) {}
    };
    const loadHistory = () => {
        if (props.saveLocation) {
            formHistory.value = LocalStorage.get<FormHistory>(saveLocationFull.value);
        }
    };

    const load = async (reload: boolean = true) => {
        if (reload) loadHistory();
        const widgets = getWidgets();
        const promises: Promise<any>[] = [];
        if (!formHistory.value) return;
        for (const widget of widgets) promises.push(widget.onLoadHistory(formHistory.value.value));
        await Promise.all(promises);
        skipSave = true;
        setValue('', formHistory.value.value);
        emit('load');
    };

    const save = debounce(_save, saveDelay.value);

    const cleanHistory = () => {
        LocalStorage.remove(saveLocationFull.value);
    };

    onMounted(() => {
        loadHistory();
        if (props.autoLoad) load(false);
    });

    return { save, load, cleanHistory, setValue, history: formHistory };
};
